"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseWith = exports.transform = exports.parseExpression = void 0;
const errorBuilder_1 = require("../errorBuilder");
const parseExpression = (expression) => {
    const tokens = tokenize(expression);
    let index = 0;
    function tokenize(expression) {
        return expression
            .replace(/\(/g, " ( ")
            .replace(/\)/g, " ) ")
            .split(" ")
            .filter((token) => token.trim().length > 0);
    }
    function parse() {
        let left = parseTerm();
        while (index < tokens.length && tokens[index] !== ")") {
            const operator = tokens[index++];
            let right;
            if (operator === "(") {
                right = parse();
            }
            else {
                right = parseTerm();
            }
            left = {
                type: operator,
                left,
                right,
            };
        }
        if (index < tokens.length && tokens[index] === ")") {
            index++;
        }
        return left;
    }
    function parseTerm() {
        const token = tokens[index++];
        if (token === "(") {
            return parse();
        }
        else {
            const operator = tokens[index++];
            let tokenIndex = index;
            let value = tokens[tokenIndex];
            while (tokenIndex < tokens.length) {
                if (value[0] !== "\"" && value[0] !== "[") {
                    break;
                }
                if (value[0] === "\"" && value.trim()[value.trim().length - 1] === "\"") {
                    break;
                }
                if (value[0] === "[" && value.trim()[value.trim().length - 1] === "]") {
                    break;
                }
                tokenIndex++;
                const currentToken = tokens[tokenIndex];
                value += " " + currentToken;
            }
            index = tokenIndex + 1;
            return {
                type: operator,
                left: {
                    type: "property",
                    name: token,
                },
                right: {
                    type: "literal",
                    value: JSON.parse(value.trim()), // Loại bỏ khoảng trắng thừa
                },
            };
        }
    }
    return parse();
};
exports.parseExpression = parseExpression;
const transform = function (type, left, right) {
    const result = {};
    const typeQueries = ['eq', 'ne', 'lt', 'lte', 'gt', 'gte', 'search', 'in', 'nin', 'and', 'or'];
    if (!typeQueries.includes(type)) {
        throw errorBuilder_1.ErrorBuilder.validator(`The type ${type} is not supported`);
    }
    if (type === 'eq' && right.type === 'literal') {
        _prop(result, left, { $eq: right.value });
    }
    if (type === 'ne' && right.type === 'literal') {
        _prop(result, left, { $ne: right.value });
    }
    if (type === 'lt' && right.type === 'literal') {
        _prop(result, left, { $lt: right.value });
    }
    if (type === 'gt' && right.type === 'literal') {
        _prop(result, left, { $gt: right.value });
    }
    if (type === 'lte' && right.type === 'literal') {
        _prop(result, left, { $lte: right.value });
    }
    if (type === 'gte' && right.type === 'literal') {
        _prop(result, left, { $gte: right.value });
    }
    if (type === 'search' && right.type === 'literal') {
        _prop(result, left, { $regex: right.value, $options: "i" });
    }
    if (type === 'in' && right.type === 'literal') {
        if (!Array.isArray(right.value)) {
            throw errorBuilder_1.ErrorBuilder.validator(`The ${left.name} must be an array`);
        }
        _prop(result, left, { $in: right.value });
    }
    if (type === 'nin' && right.type === 'literal') {
        if (!Array.isArray(right.value)) {
            throw errorBuilder_1.ErrorBuilder.validator(`The ${left.name} must be an array`);
        }
        _prop(result, left, { $nin: right.value });
    }
    if (type === 'and') {
        result.$and = result.$and || [];
        result.$and.push((0, exports.transform)(left.type, left.left, left.right));
        result.$and.push((0, exports.transform)(right.type, right.left, right.right));
    }
    if (type === 'or') {
        result.$or = result.$or || [];
        result.$or.push((0, exports.transform)(left.type, left.left, left.right));
        result.$or.push((0, exports.transform)(right.type, right.left, right.right));
    }
    // if (type === 'functioncall') {
    //   switch (func) {
    //     case 'substringof': substringof(this, result)
    //   }
    // }
    function _prop(result, left, rightValue) {
        if (left.type === 'property' && left.name.indexOf('/') !== -1) {
            const fragments = left.name.split('/');
            const obj = result[fragments[0]] || {};
            for (let i = 1; i < fragments.length; i++) {
                if (i === (fragments.length - 1)) {
                    obj[fragments[i]] = rightValue;
                }
                else {
                    obj[fragments[i]] = obj[fragments[i]] || {};
                }
            }
            result[fragments[0]] = obj;
        }
        else {
            result[left.name] = rightValue;
        }
    }
    return result;
};
exports.transform = transform;
function countBracket(str, char) {
    let count = 0;
    let escaped = false;
    for (let i = 0; i < str.length; i++) {
        if (str[i] === "\"") {
            escaped = !escaped;
        }
        else if (str[i] === char && !escaped) {
            count++;
        }
    }
    return count;
}
function getFirstOperator(str) {
    const inputFilter = "$filter";
    const regexFilter = new RegExp(`^${inputFilter.replace(/[-\/\\^$*+?.()|[\]{}]/g, "\\$&")}`);
    const isFilterAtBeginning = regexFilter.test(str.trim());
    if (isFilterAtBeginning) {
        return "filter";
    }
    const inputWith = "$with";
    const regexWith = new RegExp(`^${inputWith.replace(/[-\/\\^$*+?.()|[\]{}]/g, "\\$&")}`);
    const isWithAtBeginning = regexWith.test(str.trim());
    if (isWithAtBeginning) {
        return "with";
    }
    return null;
}
function parseWith(withParam) {
    const expand = [];
    const tokens = withParam.split(',');
    let countOpenRemaining = 0;
    let valueModel = "";
    for (const token of tokens) {
        valueModel = valueModel + "," + token;
        const expandItem = { model: valueModel.trim().replace(/^,|,$/g, ''), filter: undefined, with: undefined };
        countOpenRemaining = +countBracket(expandItem.model, "(") - countBracket(expandItem.model, ")");
        if (countOpenRemaining == 0) {
            if (countBracket(expandItem.model, "(")) {
                const withSingle = parseWithSingle(expandItem.model);
                withSingle.model = withSingle.model.trim();
                withSingle.filter = withSingle.filter ? withSingle.filter.trim() : undefined;
                expand.push(withSingle);
            }
            else {
                expandItem.model = expandItem.model.trim();
                expandItem.filter = expandItem.filter ? expandItem.filter.trim() : undefined;
                expand.push(expandItem);
            }
            valueModel = "";
        }
        else if (countOpenRemaining > 0) {
            continue;
        }
        else {
            throw errorBuilder_1.ErrorBuilder.validator("The operator with is invalid");
        }
    }
    function parseWithSingle(withParam) {
        const expandItem = { model: '', filter: undefined, with: undefined };
        const start = withParam.indexOf("(");
        const end = withParam.lastIndexOf(")");
        if (start !== -1 && end !== -1) {
            expandItem.model = withParam.substring(0, start);
            let strWithAndFilter = withParam.substring(expandItem.model.length + 1, end);
            let countOpenRemaining = 0;
            let value = "";
            if (strWithAndFilter.includes("$with") || strWithAndFilter.includes("$filter")) {
                let tokens = strWithAndFilter.split(',');
                for (let i = 0; i < tokens.length; i++) {
                    const token = tokens[i];
                    value = value ? (value + "," + token) : token;
                    countOpenRemaining = +countBracket(value, "(") - countBracket(value, ")");
                    if (countOpenRemaining == 0) {
                        const firstOperator = getFirstOperator(value);
                        if (!firstOperator) {
                            throw errorBuilder_1.ErrorBuilder.validator("The operator with is invalid");
                        }
                        if (firstOperator == "with") {
                            if (i < tokens.length - 1 && tokens[i + 1][0] != ")" && tokens[i + 1].trim()[0] != "$") {
                                continue;
                            }
                            expandItem[firstOperator] = parseWith(value.substring(firstOperator.length + 2, value.length));
                        }
                        else {
                            if (i < value.length - 1) {
                                const valueFilter = value;
                                let countQuotationMark = 0;
                                for (let i = 0; i < valueFilter.length; i++) {
                                    if (valueFilter[i] === "\"") {
                                        countQuotationMark++;
                                    }
                                }
                                if (countQuotationMark % 2) {
                                    continue;
                                }
                            }
                            expandItem[firstOperator] = value.substring(firstOperator.length + 2, value.length);
                        }
                        break;
                    }
                    else if (countOpenRemaining > 0) {
                        continue;
                    }
                    else {
                        throw errorBuilder_1.ErrorBuilder.validator("The operator with is invalid");
                    }
                }
            }
            strWithAndFilter = strWithAndFilter.substring(value.length + 1, strWithAndFilter.length).trim();
            value = "";
            if (strWithAndFilter.includes("$with") || strWithAndFilter.includes("$filter")) {
                const tokens = strWithAndFilter.split(',');
                countOpenRemaining = 0;
                for (let i = 0; i < tokens.length; i++) {
                    const token = tokens[i];
                    value = value ? (value + "," + token) : token;
                    countOpenRemaining = +countBracket(value, "(") - countBracket(value, ")");
                    if (countOpenRemaining == 0) {
                        const firstOperator = getFirstOperator(value);
                        if (!firstOperator) {
                            throw errorBuilder_1.ErrorBuilder.validator("The operator with is invalid");
                        }
                        if (firstOperator == "with") {
                            if (i < tokens.length - 1 && tokens[i + 1][0] != ")" && tokens[i + 1].trim()[0] != "$") {
                                continue;
                            }
                            expandItem[firstOperator] = parseWith(value.substring(firstOperator.length + 2, value.length));
                        }
                        else {
                            const valueFilter = value;
                            let countQuotationMark = 0;
                            for (let i = 0; i < valueFilter.length; i++) {
                                if (valueFilter[i] === "\"") {
                                    countQuotationMark++;
                                }
                            }
                            if (countQuotationMark % 2) {
                                continue;
                            }
                            expandItem[firstOperator] = value.substring(firstOperator.length + 2, value.length);
                        }
                        break;
                    }
                    else if (countOpenRemaining > 0) {
                        continue;
                    }
                    else {
                        throw errorBuilder_1.ErrorBuilder.validator("The operator with is invalid");
                    }
                }
            }
        }
        else {
            expandItem.model = withParam.trim();
        }
        return expandItem;
    }
    return expand;
}
exports.parseWith = parseWith;
